@@ -23,21 +23,22 @@ import warnings
 from oslo_concurrency import processutils
 from oslo_config import cfg
 from oslo_utils import excutils
-from oslo_utils import imageutils
-from oslo_utils import units
-import tenacity
 
 from ironic_lib.common.i18n import _
 from ironic_lib import disk_partitioner
 from ironic_lib import exception
+from ironic_lib import qemu_img
 from ironic_lib import utils
 
 
 opts = [
     cfg.IntOpt('efi_system_partition_size',
-               default=200,
+               default=550,
                help='Size of EFI system partition in MiB when configuring '
-                    'UEFI systems for local boot.'),
+                    'UEFI systems for local boot. A common minimum is ~200 '
+                    'megabytes, however OS driven firmware updates and '
+                    'unikernel usage generally requires more space on the '
+                    'efi partition.'),
     cfg.IntOpt('bios_boot_partition_size',
                default=1,
                help='Size of BIOS Boot partition in MiB when configuring '
@@ -53,13 +54,6 @@ opts = [
                default=10,
                help='Maximum number of attempts to try to read the '
                     'partition.'),
-    cfg.IntOpt('image_convert_memory_limit',
-               default=2048,
-               help='Memory limit for "qemu-img convert" in MiB. Implemented '
-                    'via the address space resource limit.'),
-    cfg.IntOpt('image_convert_attempts',
-               default=3,
-               help='Number of attempts to convert an image.'),
 ]
 
 CONF = cfg.CONF
@@ -75,22 +69,12 @@ _PARTED_TABLE_TYPE_RE = re.compile(r'^.*partition\s+table\s*:\s*(gpt|msdos)',
 CONFIGDRIVE_LABEL = "config-2"
 MAX_CONFIG_DRIVE_SIZE_MB = 64
 
-GPT_SIZE_SECTORS = 33
-
 # Maximum disk size supported by MBR is 2TB (2 * 1024 * 1024 MB)
 MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152
 
-# Limit the memory address space to 1 GiB when running qemu-img
-QEMU_IMG_LIMITS = None
-
-
-def _qemu_img_limits():
-    global QEMU_IMG_LIMITS
-    if QEMU_IMG_LIMITS is None:
-        QEMU_IMG_LIMITS = processutils.ProcessLimits(
-            address_space=CONF.disk_utils.image_convert_memory_limit
-            * units.Mi)
-    return QEMU_IMG_LIMITS
+# Backward compatibility, do not use
+qemu_img_info = qemu_img.image_info
+convert_image = qemu_img.convert_image
 
 
 def list_partitions(device):
@@ -149,7 +133,7 @@ def count_mbr_partitions(device):
     output = output.replace('<', '').replace('>', '')
     partitions = [int(s) for s in output.split() if s.isdigit()]
 
-    return(sum(i < 5 for i in partitions), sum(i > 4 for i in partitions))
+    return (sum(i < 5 for i in partitions), sum(i > 4 for i in partitions))
 
 
 def get_disk_identifier(dev):
@@ -482,73 +466,12 @@ def dd(src, dst, conv_flags=None):
              *extra_args)
 
 
-def qemu_img_info(path):
-    """Return an object containing the parsed output from qemu-img info."""
-    if not os.path.exists(path):
-        return imageutils.QemuImgInfo()
-
-    out, err = utils.execute('env', 'LC_ALL=C', 'LANG=C',
-                             'qemu-img', 'info', path,
-                             '--output=json',
-                             prlimit=_qemu_img_limits())
-    return imageutils.QemuImgInfo(out, format='json')
-
-
-def _retry_on_res_temp_unavailable(exc):
-    if (isinstance(exc, processutils.ProcessExecutionError)
-            and ('Resource temporarily unavailable' in exc.stderr
-                 or 'Cannot allocate memory' in exc.stderr)):
-        return True
-    return False
-
-
-@tenacity.retry(
-    retry=tenacity.retry_if_exception(_retry_on_res_temp_unavailable),
-    stop=tenacity.stop_after_attempt(CONF.disk_utils.image_convert_attempts),
-    reraise=True)
-def convert_image(source, dest, out_format, run_as_root=False, cache=None,
-                  out_of_order=False, sparse_size=None):
-    """Convert image to other format."""
-    cmd = ['qemu-img', 'convert', '-O', out_format]
-    if cache is not None:
-        cmd += ['-t', cache]
-    if sparse_size is not None:
-        cmd += ['-S', sparse_size]
-    if out_of_order:
-        cmd.append('-W')
-    cmd += [source, dest]
-    # NOTE(TheJulia): Staticly set the MALLOC_ARENA_MAX to prevent leaking
-    # and the creation of new malloc arenas which will consume the system
-    # memory. If limited to 1, qemu-img consumes ~250 MB of RAM, but when
-    # another thread tries to access a locked section of memory in use with
-    # another thread, then by default a new malloc arena is created,
-    # which essentially balloons the memory requirement of the machine.
-    # Default for qemu-img is 8 * nCPU * ~250MB (based on defaults +
-    # thread/code/process/library overhead. In other words, 64 GB. Limiting
-    # this to 3 keeps the memory utilization in happy cases below the overall
-    # threshold which is in place in case a malicious image is attempted to
-    # be passed through qemu-img.
-    env_vars = {'MALLOC_ARENA_MAX': '3'}
-    try:
-        utils.execute(*cmd, run_as_root=run_as_root,
-                      prlimit=_qemu_img_limits(),
-                      use_standard_locale=True,
-                      env_variables=env_vars)
-    except processutils.ProcessExecutionError as e:
-        if ('Resource temporarily unavailable' in e.stderr
-            or 'Cannot allocate memory' in e.stderr):
-            LOG.debug('Failed to convert image, retrying. Error: %s', e)
-            # Sync disk caches before the next attempt
-            utils.execute('sync')
-        raise
-
-
 def populate_image(src, dst, conv_flags=None):
-    data = qemu_img_info(src)
+    data = qemu_img.image_info(src)
     if data.file_format == 'raw':
         dd(src, dst, conv_flags=conv_flags)
     else:
-        convert_image(src, dst, 'raw', True, sparse_size='0')
+        qemu_img.convert_image(src, dst, 'raw', True, sparse_size='0')
 
 
 def block_uuid(dev):
@@ -572,7 +495,7 @@ def get_image_mb(image_path, virtual_size=True):
     if not virtual_size:
         image_byte = os.path.getsize(image_path)
     else:
-        data = qemu_img_info(image_path)
+        data = qemu_img.image_info(image_path)
         image_byte = data.virtual_size
 
     # round up size to MB
@@ -587,6 +510,20 @@ def get_dev_block_size(dev):
     return int(block_sz)
 
 
+def get_dev_byte_size(dev):
+    """Get the device size in bytes."""
+    byte_sz, cmderr = utils.execute('blockdev', '--getsize64', dev,
+                                    run_as_root=True)
+    return int(byte_sz)
+
+
+def get_dev_sector_size(dev):
+    """Get the device logical sector size in bytes."""
+    sect_sz, cmderr = utils.execute('blockdev', '--getss', dev,
+                                    run_as_root=True)
+    return int(sect_sz)
+
+
 def destroy_disk_metadata(dev, node_uuid):
     """Destroy metadata structures on node's disk.
 
@@ -620,22 +557,34 @@ def destroy_disk_metadata(dev, node_uuid):
     # This is the same bug as
     # https://bugs.launchpad.net/ironic-python-agent/+bug/1737556
 
+    sector_size = get_dev_sector_size(dev)
+    # https://uefi.org/specs/UEFI/2.10/05_GUID_Partition_Table_Format.html If
+    # the block size is 512, the First Usable LBA must be greater than or equal
+    # to 34 [...] if the logical block size is 4096, the First Usable LBA must
+    # be greater than or equal to 6
+    if sector_size == 512:
+        gpt_sectors = 33
+    elif sector_size == 4096:
+        gpt_sectors = 5
+
     # Overwrite the Primary GPT, catch very small partitions (like EBRs)
+    dd_bs = 'bs=%s' % sector_size
     dd_device = 'of=%s' % dev
-    dd_count = 'count=%s' % GPT_SIZE_SECTORS
-    dev_size = get_dev_block_size(dev)
-    if dev_size < GPT_SIZE_SECTORS:
-        dd_count = 'count=%s' % dev_size
-    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
-                  run_as_root=True, use_standard_locale=True)
+    dd_count = 'count=%s' % gpt_sectors
+    dev_size = get_dev_byte_size(dev)
+    if dev_size < gpt_sectors * sector_size:
+        dd_count = 'count=%s' % int(dev_size / sector_size)
+    utils.execute('dd', dd_bs, 'if=/dev/zero', dd_device, dd_count,
+                  'oflag=direct', run_as_root=True, use_standard_locale=True)
 
     # Overwrite the Secondary GPT, do this only if there could be one
-    if dev_size > GPT_SIZE_SECTORS:
-        gpt_backup = dev_size - GPT_SIZE_SECTORS
+    if dev_size > gpt_sectors * sector_size:
+        gpt_backup = int(dev_size / sector_size - gpt_sectors)
         dd_seek = 'seek=%i' % gpt_backup
-        dd_count = 'count=%s' % GPT_SIZE_SECTORS
-        utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
-                      dd_seek, run_as_root=True, use_standard_locale=True)
+        dd_count = 'count=%s' % gpt_sectors
+        utils.execute('dd', dd_bs, 'if=/dev/zero', dd_device, dd_count,
+                      'oflag=direct', dd_seek, run_as_root=True,
+                      use_standard_locale=True)
 
     # Go ahead and let sgdisk run as well.
     utils.execute('sgdisk', '-Z', dev, run_as_root=True,
@@ -725,7 +674,7 @@ def udev_settle():
 def partprobe(device, attempts=None):
     """Probe partitions on the given device.
 
-    :param device: The block device containing paritions that is attempting
+    :param device: The block device containing partitions that is attempting
                    to be updated.
     :param attempts: Number of attempts to run partprobe, the default is read
                      from the configuration.
@@ -762,7 +711,7 @@ def trigger_device_rescan(device, attempts=None):
     When ramdisks are not in use, this also helps ensure that data has
     been safely flushed across the wire, such as on iscsi connections.
 
-    :param device: The block device containing paritions that is attempting
+    :param device: The block device containing partitions that is attempting
                    to be updated.
     :param attempts: Number of attempts to run partprobe, the default is read
                      from the configuration.
@@ -774,6 +723,7 @@ def trigger_device_rescan(device, attempts=None):
     # kernel.
     udev_settle()
     partprobe(device, attempts=attempts)
+    udev_settle()
     try:
         # Also verify that the partitioning is correct now.
         utils.execute('sgdisk', '-v', device, run_as_root=True)
