@@ -18,12 +18,16 @@
 
 """Utilities and helper functions."""
 
+import contextlib
 import copy
 import errno
 import ipaddress
 import logging
 import os
 import re
+import shlex
+import shutil
+import tempfile
 from urllib import parse as urlparse
 
 from oslo_concurrency import processutils
@@ -60,34 +64,28 @@ VALID_ROOT_DEVICE_HINTS = {
 ROOT_DEVICE_HINTS_GRAMMAR = specs_matcher.make_grammar()
 
 
-def execute(*cmd, **kwargs):
+def execute(*cmd, use_standard_locale=False, log_stdout=True, **kwargs):
     """Convenience wrapper around oslo's execute() method.
 
     Executes and logs results from a system command. See docs for
     oslo_concurrency.processutils.execute for usage.
 
     :param cmd: positional arguments to pass to processutils.execute()
-    :param use_standard_locale: keyword-only argument. True | False.
-                                Defaults to False. If set to True,
+    :param use_standard_locale: Defaults to False. If set to True,
                                 execute command with standard locale
                                 added to environment variables.
-    :param log_stdout: keyword-only argument. True | False. Defaults
-                       to True. If set to True, logs the output.
+    :param log_stdout: Defaults to True. If set to True, logs the output.
     :param kwargs: keyword arguments to pass to processutils.execute()
     :returns: (stdout, stderr) from process execution
     :raises: UnknownArgumentError on receiving unknown arguments
     :raises: ProcessExecutionError
     :raises: OSError
     """
-
-    use_standard_locale = kwargs.pop('use_standard_locale', False)
     if use_standard_locale:
         env = kwargs.pop('env_variables', os.environ.copy())
         env['LC_ALL'] = 'C'
         kwargs['env_variables'] = env
 
-    log_stdout = kwargs.pop('log_stdout', True)
-
     # If root_helper config is not specified, no commands are run as root.
     run_as_root = kwargs.get('run_as_root', False)
     if run_as_root:
@@ -96,13 +94,34 @@ def execute(*cmd, **kwargs):
         else:
             kwargs['root_helper'] = CONF.ironic_lib.root_helper
 
-    result = processutils.execute(*cmd, **kwargs)
-    LOG.debug('Execution completed, command line is "%s"',
-              ' '.join(map(str, cmd)))
-    if log_stdout:
-        LOG.debug('Command stdout is: "%s"', result[0])
-    LOG.debug('Command stderr is: "%s"', result[1])
-    return result
+    def _log(stdout, stderr):
+        if log_stdout:
+            try:
+                LOG.debug('Command stdout is: "%s"', stdout)
+            except UnicodeEncodeError:
+                LOG.debug('stdout contains invalid UTF-8 characters')
+                stdout = (stdout.encode('utf8', 'surrogateescape')
+                          .decode('utf8', 'ignore'))
+                LOG.debug('Command stdout is: "%s"', stdout)
+        try:
+            LOG.debug('Command stderr is: "%s"', stderr)
+        except UnicodeEncodeError:
+            LOG.debug('stderr contains invalid UTF-8 characters')
+            stderr = (stderr.encode('utf8', 'surrogateescape')
+                      .decode('utf8', 'ignore'))
+            LOG.debug('Command stderr is: "%s"', stderr)
+
+    try:
+        result = processutils.execute(*cmd, **kwargs)
+    except FileNotFoundError:
+        with excutils.save_and_reraise_exception():
+            LOG.debug('Command not found: "%s"', ' '.join(map(str, cmd)))
+    except processutils.ProcessExecutionError as exc:
+        with excutils.save_and_reraise_exception():
+            _log(exc.stdout, exc.stderr)
+    else:
+        _log(result[0], result[1])
+        return result
 
 
 def try_execute(*cmd, **kwargs):
@@ -185,7 +204,7 @@ def dd(src, dst, *args):
     """
     LOG.debug("Starting dd process.")
     execute('dd', 'if=%s' % src, 'of=%s' % dst, *args,
-            use_standard_locale=True, run_as_root=True, check_exit_code=[0])
+            use_standard_locale=True, run_as_root=True)
 
 
 def is_http_url(url):
@@ -585,3 +604,76 @@ def get_route_source(dest, ignore_link_local=True):
     except (IndexError, ValueError):
         LOG.debug('No route to host %(dest)s, route record: %(rec)s',
                   {'dest': dest, 'rec': out})
+
+
+@contextlib.contextmanager
+def mounted(source, dest=None, opts=None, fs_type=None,
+            mount_attempts=1, umount_attempts=3):
+    """A context manager for a temporary mount.
+
+    :param source: A device to mount.
+    :param dest: Mount destination. If not specified, a temporary directory
+        will be created and removed afterwards. An existing destination is
+        not removed.
+    :param opts: Mount options (``-o`` argument).
+    :param fs_type: File system type (``-t`` argument).
+    :param mount_attempts: A number of attempts to mount the device.
+    :param umount_attempts: A number of attempts to unmount the device.
+    :returns: A generator yielding the destination.
+    """
+    params = []
+    if opts:
+        params.extend(['-o', ','.join(opts)])
+    if fs_type:
+        params.extend(['-t', fs_type])
+
+    if dest is None:
+        dest = tempfile.mkdtemp()
+        clean_up = True
+    else:
+        clean_up = False
+
+    mounted = False
+    try:
+        execute("mount", source, dest, *params, run_as_root=True,
+                attempts=mount_attempts, delay_on_retry=True)
+        mounted = True
+        yield dest
+    finally:
+        if mounted:
+            try:
+                execute("umount", dest, run_as_root=True,
+                        attempts=umount_attempts, delay_on_retry=True)
+            except (EnvironmentError,
+                    processutils.ProcessExecutionError) as exc:
+                LOG.warning(
+                    'Unable to unmount temporary location %(dest)s: %(err)s',
+                    {'dest': dest, 'err': exc})
+                # NOTE(dtantsur): don't try to remove a still mounted location
+                clean_up = False
+
+        if clean_up:
+            try:
+                shutil.rmtree(dest)
+            except EnvironmentError as exc:
+                LOG.warning(
+                    'Unable to remove temporary location %(dest)s: %(err)s',
+                    {'dest': dest, 'err': exc})
+
+
+def parse_device_tags(output):
+    """Parse tags from the lsblk/blkid output.
+
+    Parses format KEY="VALUE" KEY2="VALUE2".
+
+    :return: a generator yielding dicts with information from each line.
+    """
+    for line in output.strip().split('\n'):
+        if line.strip():
+            try:
+                yield {key: value for key, value in
+                       (v.split('=', 1) for v in shlex.split(line))}
+            except ValueError as err:
+                raise ValueError(
+                    _("Malformed blkid/lsblk output line '%(line)s': %(err)s")
+                    % {'line': line, 'err': err})
